"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.injectedClientEntries = void 0;
var _querystring = require("querystring");
var _path = _interopRequireDefault(require("path"));
var _webpack = require("next/dist/compiled/webpack/webpack");
var _utils = require("../loaders/utils");
var _normalizePagePath = require("../../../shared/lib/page-path/normalize-page-path");
var _denormalizePagePath = require("../../../shared/lib/page-path/denormalize-page-path");
var _onDemandEntryHandler = require("../../../server/dev/on-demand-entry-handler");
var _constants = require("../../../lib/constants");
var _constants1 = require("../../../shared/lib/constants");
function _interopRequireDefault(obj) {
    return obj && obj.__esModule ? obj : {
        default: obj
    };
}
const PLUGIN_NAME = "ClientEntryPlugin";
const injectedClientEntries = new Map();
exports.injectedClientEntries = injectedClientEntries;
// TODO-APP: ensure .scss / .sass also works.
const regexCSS = /\.css$/;
// TODO-APP: move CSS manifest generation to the flight manifest plugin.
const flightCSSManifest = {};
class FlightClientEntryPlugin {
    constructor(options){
        this.dev = options.dev;
        this.isEdgeServer = options.isEdgeServer;
    }
    apply(compiler) {
        compiler.hooks.compilation.tap(PLUGIN_NAME, (compilation, { normalModuleFactory  })=>{
            compilation.dependencyFactories.set(_webpack.webpack.dependencies.ModuleDependency, normalModuleFactory);
            compilation.dependencyTemplates.set(_webpack.webpack.dependencies.ModuleDependency, new _webpack.webpack.dependencies.NullDependency.Template());
        });
        compiler.hooks.finishMake.tapPromise(PLUGIN_NAME, (compilation)=>{
            return this.createClientEndpoints(compiler, compilation);
        });
    }
    async createClientEndpoints(compiler, compilation) {
        const promises = [];
        // For each SC server compilation entry, we need to create its corresponding
        // client component entry.
        for (const [name, entry] of compilation.entries.entries()){
            var ref;
            // If the request is not for `app` directory entry skip it.
            // if (!entry.request || !entry.request.startsWith('next-app-loader')) {
            //   continue
            // }
            // Check if the page entry is a server component or not.
            const entryDependency = (ref = entry.dependencies) == null ? void 0 : ref[0];
            // Ensure only next-app-loader entries are handled.
            if (!entryDependency || !entryDependency.request || !entryDependency.request.startsWith("next-app-loader?")) {
                continue;
            }
            // TODO-APP: create client-side entrypoint per layout/page.
            // const entryModule: webpack5.NormalModule =
            //   compilation.moduleGraph.getResolvedModule(entryDependency)
            // for (const connection of compilation.moduleGraph.getOutgoingConnections(
            //   entryModule
            // )) {
            //   const layoutOrPageDependency = connection.dependency
            //   // const layoutOrPageRequest = connection.dependency.request
            //   const [clientComponentImports, cssImports] =
            //     this.collectClientComponentsAndCSSForDependency(
            //       compiler.context,
            //       compilation,
            //       layoutOrPageDependency
            //     )
            //   Object.assign(serverCSSManifest, cssImports)
            //   promises.push(
            //     this.injectClientEntryAndSSRModules(
            //       compiler,
            //       compilation,
            //       name,
            //       entryDependency,
            //       clientComponentImports
            //     )
            //   )
            // }
            const [clientComponentImports, cssImports] = this.collectClientComponentsAndCSSForDependency(compiler.context, compilation, entryDependency);
            Object.assign(flightCSSManifest, cssImports);
            promises.push(this.injectClientEntryAndSSRModules(compiler, compilation, name, entryDependency, clientComponentImports));
        }
        compilation.hooks.processAssets.tap({
            name: PLUGIN_NAME,
            // Have to be in the optimize stage to run after updating the CSS
            // asset hash via extract mini css plugin.
            // @ts-ignore TODO: Remove ignore when webpack 5 is stable
            stage: _webpack.webpack.Compilation.PROCESS_ASSETS_STAGE_OPTIMIZE_HASH
        }, (assets)=>{
            assets[_constants1.FLIGHT_SERVER_CSS_MANIFEST + ".json"] = new _webpack.sources.RawSource(JSON.stringify(flightCSSManifest));
        });
        const res = await Promise.all(promises);
        // Invalidate in development to trigger recompilation
        const invalidator = (0, _onDemandEntryHandler).getInvalidator();
        // Check if any of the entry injections need an invalidation
        if (invalidator && res.includes(true)) {
            invalidator.invalidate([
                _constants1.COMPILER_NAMES.client
            ]);
        }
    }
    collectClientComponentsAndCSSForDependency(context, compilation, dependency /* Dependency */ ) {
        /**
     * Keep track of checked modules to avoid infinite loops with recursive imports.
     */ const visitedBySegment = {};
        const clientComponentImports = [];
        const serverCSSImports = {};
        const filterClientComponents = (dependencyToFilter, segmentPath)=>{
            var ref;
            const mod = compilation.moduleGraph.getResolvedModule(dependencyToFilter);
            if (!mod) return;
            // Keep client imports as simple
            // native or installed js module: -> raw request, e.g. next/head
            // client js or css: -> user request
            const rawRequest = mod.rawRequest;
            // Request could be undefined or ''
            if (!rawRequest) return;
            const modRequest = !rawRequest.endsWith(".css") && !rawRequest.startsWith(".") && !rawRequest.startsWith("/") && !rawRequest.startsWith(_constants.APP_DIR_ALIAS) ? rawRequest : (ref = mod.resourceResolveData) == null ? void 0 : ref.path;
            // Ensure module is not walked again if it's already been visited
            if (!visitedBySegment[segmentPath]) {
                visitedBySegment[segmentPath] = new Set();
            }
            if (!modRequest || visitedBySegment[segmentPath].has(modRequest)) return;
            visitedBySegment[segmentPath].add(modRequest);
            const isLayoutOrPage = /\/(layout|page)(\.server|\.client)?\.(js|ts)x?$/.test(modRequest);
            const isCSS = regexCSS.test(modRequest);
            const isClientComponent = _utils.clientComponentRegex.test(modRequest);
            if (isCSS) {
                serverCSSImports[segmentPath] = serverCSSImports[segmentPath] || [];
                serverCSSImports[segmentPath].push(modRequest);
            }
            // Check if request is for css file.
            if (isClientComponent || isCSS) {
                clientComponentImports.push(modRequest);
                return;
            }
            if (isLayoutOrPage) {
                segmentPath = _path.default.relative(_path.default.join(context, "app"), _path.default.dirname(modRequest)).replace(/\\/g, "/");
                if (segmentPath !== "") {
                    segmentPath = "/" + segmentPath;
                }
                // If it's a page, add an extra '/' to the segments
                if (/\/(page)(\.server|\.client)?\.(js|ts)x?$/.test(modRequest)) {
                    segmentPath += "/";
                }
            }
            compilation.moduleGraph.getOutgoingConnections(mod).forEach((connection)=>{
                filterClientComponents(connection.dependency, segmentPath);
            });
        };
        // Traverse the module graph to find all client components.
        filterClientComponents(dependency, "");
        return [
            clientComponentImports,
            serverCSSImports
        ];
    }
    async injectClientEntryAndSSRModules(compiler, compilation, entryName, entryDependency, clientComponentImports) {
        let shouldInvalidate = false;
        const entryModule = compilation.moduleGraph.getResolvedModule(entryDependency);
        const routeInfo = entryModule.buildInfo.route || {
            page: (0, _denormalizePagePath).denormalizePagePath(entryName.replace(/^pages/, "")),
            absolutePagePath: entryModule.resource
        };
        const loaderOptions = {
            modules: clientComponentImports,
            server: false
        };
        const clientLoader = `next-flight-client-entry-loader?${(0, _querystring).stringify(loaderOptions)}!`;
        const clientSSRLoader = `next-flight-client-entry-loader?${(0, _querystring).stringify({
            ...loaderOptions,
            server: true
        })}!`;
        const bundlePath = "app" + (0, _normalizePagePath).normalizePagePath(routeInfo.page);
        // Add for the client compilation
        // Inject the entry to the client compiler.
        if (this.dev) {
            const pageKey = _constants1.COMPILER_NAMES.client + routeInfo.page;
            if (!_onDemandEntryHandler.entries[pageKey]) {
                _onDemandEntryHandler.entries[pageKey] = {
                    type: _onDemandEntryHandler.EntryTypes.CHILD_ENTRY,
                    parentEntries: new Set([
                        entryName
                    ]),
                    bundlePath,
                    // absolutePagePath: routeInfo.absolutePagePath,
                    request: clientLoader,
                    dispose: false,
                    lastActiveTime: Date.now()
                };
                shouldInvalidate = true;
            } else {
                const entryData = _onDemandEntryHandler.entries[pageKey];
                // New version of the client loader
                if (entryData.request !== clientLoader) {
                    entryData.request = clientLoader;
                    shouldInvalidate = true;
                }
                if (entryData.type === _onDemandEntryHandler.EntryTypes.CHILD_ENTRY) {
                    entryData.parentEntries.add(entryName);
                }
            }
        } else {
            injectedClientEntries.set(bundlePath, clientLoader);
        }
        // Inject the entry to the server compiler (__sc_client__).
        const clientComponentEntryDep = _webpack.webpack.EntryPlugin.createDependency(clientSSRLoader, {
            name: bundlePath
        });
        // Add the dependency to the server compiler.
        await this.addEntry(compilation, // Reuse compilation context.
        compiler.context, clientComponentEntryDep, {
            // By using the same entry name
            name: entryName,
            // Layer should be undefined for the SSR modules
            // This ensures the client components are
            layer: undefined
        });
        return shouldInvalidate;
    }
    addEntry(compilation, context, entry /* Dependency */ , options /* EntryOptions */ ) /* Promise<module> */ {
        return new Promise((resolve, reject)=>{
            compilation.entries.get(options.name).includeDependencies.push(entry);
            compilation.hooks.addEntry.call(entry, options);
            compilation.addModuleTree({
                context,
                dependency: entry,
                contextInfo: {
                    issuerLayer: options.layer
                }
            }, (err, module)=>{
                if (err) {
                    compilation.hooks.failedEntry.call(entry, options, err);
                    return reject(err);
                }
                compilation.hooks.succeedEntry.call(entry, options, module);
                return resolve(module);
            });
        });
    }
}
exports.FlightClientEntryPlugin = FlightClientEntryPlugin;

//# sourceMappingURL=flight-client-entry-plugin.js.map